<?php

/**
 * API response class
 *
 * This class is responsible for formatting and sending a
 * response back to the client.
 *
 * Extends Exception, to allow us to take a short-cut out of deep call
 * hierarchies. This is really for Api_Error_Response's benefit, since
 * PHP neither supports multiple inheritance, or has an exception interface.
 */
class Api_Response extends Exception
{
	protected $type = false; /**< The object type we refer to */
	protected $title = false; /**< Something pretty and describing, usually object name or type */
	/** array [id] => [title => string, content => string] */
	protected $troubleshooting = array(
		'ssl' => array(
			'title' => "SSL certificate is invalid",
			'content' => "<p>By default, op5 Monitor ships with an unsigned SSL certificate. If you do not replace this with a signed certificate, the most HTTP access methods (including the ones given in the examples) will deny your access. This is solved the following way:</p>
				<p>cURL: Add a <code>--insecure</code> flag:</p>
				<div class='curl'>
					<pre><code>curl &lt;API endpoint here&gt; <strong>--insecure</strong> -u &lt;your username&gt;:&lt;your password&gt;</code></pre>
				</div>

				<p>Python: Adding a <code>verify=False</code> flag:</p>
				<div class='python'>
					<pre><code>requests.get('&lt;API endpoint here&gt;', auth=('&lt;your username&gt;', '&lt;your username&gt;'), <strong>verify=False</strong>)</code></pre>
				</div>

				<p>PHP with cURL: Adding a <code>CURLOPT_SSL_VERIFYPEER</code> flag:</p>
				<div class='php'>
					<pre><code>curl_setopt(\$a_handle, <strong>CURLOPT_SSL_VERIFYPEER</strong>, false);</code></pre>
				</div>"
		),
		'connect' => array(
			'title' => "Can I connect to the API in another way?",
			'content' => "Yes, all you need is a way to perform a HTTP method in your language. Study the minimal examples and take note of which HTTP method is used, which URI that's used and what data is sent to the URI."
		),
		'method' => array(
			'title' => "What's with PUT, DELETE, GET..?",
			'content' => "These are HTTP methods. There are examples of each method on the object's page."
		),
		'status_code' => array(
			'title' => "Why am I getting a bad status code, like 400?",
			'content' => "The HTTP response status codes are quite general and Monitor tries to follow the formal descriptions in every possible situation. Therefore, Wikipedia's article <a href='http://en.wikipedia.org/wiki/HTTP_status_code'>List of HTTP status codes</a> should help you get back on track; for example, it says that the code 400 indicates bad syntax of the request."
		),
		'cant_find' => array(
			'title' => "Why can't the API find my object?",
			'content' => "First, login to the Monitor GUI and see if you can access it there. If you can, make sure you ask for it correctly by urlencoding the object's name twice. For example, if your host is called <em>my/host</em>, a call for <var>/api/status/my/host/a_service</var> need to be this instead: <var>/api/status/my%252Fhost/a_service</var>"
		),
		'login_custom_auth' => array(
			'title' => "How do I choose how to login with LDAP or locally?",
			'content' => '<p>When you have more than one auth module, for example "Local" and "LDAP", you need to specify which to authenticate against. This is done with the dollar character (\'$\').</p>

			<p>Thus, this regular call:<br />
			<var>curl -u user:password https://monitor/api/status/host</var></p>

			<p>
				becomes<br/>
				<var>curl -u \'user$LDAP:password\' https://monitor/api/status/host</var><br/>
				or<br/>
				<var>curl -u \'user$Local:password\' https://monitor/api/status/host</var>
			</p>

			<p>Notice how the dollar sign (\'$\') needs quoting depending on the environment (in bash, it will always need to be quoted).</p>

			<p>The first way of calling the API:<br />
			<var>curl -u user:password https://monitor/api/status/host</var><br />
			will still work, provided that you want to authenticate against the default driver. The default driver can always be specified from within the GUI: Configure -> Auth Modules.</p>'
		),
		'custom_variables' => array(
			'title' => "How do I handle custom variables?",
			'content' => "Custom variables are just key-values stored with each Nagios object, as such, the API treats them the same way. Just make sure that each key is prepended with an underscore (_), otherwise the request will be returned with an erroneous status code."
		),
		'templates' => array(
			'title' => "Why can I create objects without all required properties set? Or, How do templates work?",
			'content' => "<p>Most objects has templates, i.e. hosts have their <var>template</var> property. There is usually a default template which will be used. If that template has a set <var>file_id</var> for example, you are not required to include a <var>file_id</var> in the definition of your host, since it will be inherited from the template. You can either rely on the default template, or provide another value of <var>template</var>.</p><p>For further configuration on the matter, please see the configuration tool inside the Monitor GUI, and read about <strong>Templates</strong> in the op5 Monitor Administrator Manual.</p>"
		),
		'resource' => array(
			'title' => "What is a 'resource'?",
			'content' => "<p>When listing objects, you are given an additional property called <em>resource</em>. This let's you know the resource's location to which you might send a POST or PUT request. Do note that the resource is not permanent, since a removed/renamed object will make the URI cease to exist.</p>
				<p>The resource is urlencoded so that it is always usable in its current form.</p>"
		),
		'unknown' => array(
			'title' => "I am experiencing problems, where should I go next?",
			'content' => "You could contact your op5 Monitor technician or visit <a href='http://www.op5.com/support/'>http://www.op5.com/support/</a> to acquire more help."
		)
	);

	/**
	 * The content of these properties will be wrapped in <a href> in the HTML output
	 */
	public $properties_to_linkup = array(
		'resource',
		'next_resource'
	);
	/**
	 * Constructor
	 * @param $body Response body
	 * @param $code HTTP response code
	 */
	function __construct($body, $code = 200)
	{
		$this->body = $body;
		$this->code = $code;
	}

	/**
	 * @param $properties array
	 * @param $title string = null
	 * @return string html
	 */
	protected function _format_properties($properties, $title = null)
	{
		$properties_as_html = '<table>
			<tr>
				<th>Parameter (<span class="required">required</span>)</th>
				<th>Type</th>
				<th>Default</th>
				<th>Description</th>
			</tr>';
		if($title) {
			$title = (string) $title;
			$properties_as_html = "<h2>Properties of <strong>$title</strong> resources</h2>".$properties_as_html;
		}
		foreach($properties as $property) {
			$required_class = 'class="required"';
			if(!$property['required']) {
				$required_class = '';
			}
			$properties_as_html .= "<tr $required_class>
				<td><var>".$property['name']."</var></td>
				<td>".$property['type']."</td>
				<td>".$property['default']."</td>
				<td>".$property['description']."</td>
			</tr>";
		}
		return $properties_as_html.'</table>';
	}

	/**
	 * @param $troubleshooting array
	 * @return array [menu => string, content => string]
	 */
	protected function _format_troubleshooting($troubleshooting)
	{
		$menu = null;
		$content = null;
		foreach($troubleshooting as $key => $item) {
			$menu .= "<li><a href='#trouble_$key'>{$item['title']}</a></li>\n";
			$content .= "<dt id='trouble_$key'>{$item['title']}</dt>
				<dd>{$item['content']}</dd>\n";
		}
		return array(
		       'menu' => "<h3>Troubleshooting</h3>
				<ul>$menu</ul>",
		       'content' => "<h3>Troubleshooting</h3>
				<dl>
					$content
				</dl>"
	       );

	}

	/**
	 * Serialize output
	 *
	 * @param $output_format string
	 * @return string
	 */
	protected function _get_adjusted_response($output_format)
	{
		switch($output_format) {
			case 'xml':
				$content_type = 'application/xml';
				$body = $this->xml_encode($this->body)->saveXML();
				break;
			case 'html':
				$content_type = 'text/html';
				$title = "op5 Monitor API - $this->title";
				$menu = null;
				$description = null;
				$properties = null;
				$trouble_menu = null;
				$trouble_content = null;
				if(is_array($this->body) && isset($this->body['menu']) && isset($this->body['body'])) {
					$body = $this->html_encode($this->body['body']);
					$menu = "<h2>".$this->body['menu']['title']."</h2><ul>";
					foreach($this->body['menu']['items'] as $name => $menu_item) {
						$menu .= sprintf(
							"<li><a href='%s'>%s</a></li>",
							$menu_item['resource'],
							mb_convert_case($name, MB_CASE_TITLE)
						);
					}
					$menu .= "</ul>";
					if(isset($this->body['description'])) {
						$description = $this->body['description'];
					}
					if(isset($this->body['properties'])) {
						$properties = $this->_format_properties($this->body['properties'], $this->title);
					}
					$troubleshooting = $this->troubleshooting;
					if(isset($this->body['troubleshooting'])) {
						$troubleshooting = array_merge($this->body['troubleshooting'], $troubleshooting);
					}
					$troubles = $this->_format_troubleshooting($troubleshooting);
					$trouble_menu = $troubles['menu'];
					$trouble_content = $troubles['content'];
				} else {
					$body = $this->html_encode($this->body);
				}
				if(0 !== strpos($body, '<!DOCTYPE')) {
					$body = "<!DOCTYPE html>
					<html><head>
					<link type='text/css' rel='stylesheet' href='/api/css/reset.css' />
					<link type='text/css' rel='stylesheet' href='/api/css/monitor.css' />
					<link rel='index' href='/api' />
					<title>$title</title>
					</head><body>
					<div id='container'>
						<div id='sidebar'>
							<div id='monitor'>op5 Monitor</div>
							<p>HTTP API</p>
							<ul>
								<li><a href='/api/help/config'>Config</a></li>
								<li><a href='/api/help/report'>Report</a></li>
								<li><a href='/api/help/status'>Status</a></li>
							</ul>
							<p>This is the op5 Monitor status API. Each Nagios object can be queried by using the following documentation.</p>
							$menu
							$trouble_menu
						</div>
						<div id='main'>
							<h1>$this->title</h1>
							$description
							$properties
							$body
							$trouble_content
							<h2>Getting help</h2>
							<p>Please visit <a href='/api/help/config'>/api/help/config</a>, <a href='/api/help/report'>/api/help/report</a> or <a href='/api/help/status'>/api/help/status</a>.</p>
						</div>
					</div>
					</body></html>";
				}
				break;
			case 'json':
			default:
				$content_type = 'application/json';
				$body = json_encode($this->body);
				break;
		}
		return $body;
	}

	/**
	 * Send a response
	 *
	 * This function takes a request object to find output format from, but
	 * in case of errors parsing the input, it needs to be optional.
	 *
	 * @param $request Api_Request = null
	 * @return string
	 */
	function send(Api_Request $request=null)
	{
		$output_format = false;
		if ($request) {
			$output_format = $request->get_output_format();
			$path = $request->get_path();
			$this->type = reset($path);
			$this->title = end($path);
		}

		$status_response = $this->code;
		switch($this->code) {
			case 200:
				$status_response .= " OK";
				break;
			case 201:
				$status_response .= " Created";
				break;
			case 400:
				$status_response .= " Bad request";
				break;
			case 401:
				$status_response .= " Unauthorized";
				break;
			case 403:
				$status_response .= " Forbidden";
				break;
			case 404:
				$status_response .= " Not Found";
				break;
			case 405:
				$status_response .= " Method Not Allowed";
				break;
			case 500:
				$status_response .= " Internal Server Error";
				break;
			case 501:
				$status_response .= " Not Implemented";
				break;
		}
		header("HTTP/1.0 $status_response");
		switch($output_format) {
			case 'xml':
				$content_type = 'application/xml';
				break;
			case 'html':
				$content_type = 'text/html';
				break;
			case 'json':
			default:
				$content_type = 'application/json';
				break;
		}
		header("Content-Type: $content_type; charset=utf-8");
		return $this->_get_adjusted_response($output_format);
	}

	private function xml_encode_worker($k, $v, $parent, $doc, $elem_name="value") {
		if (!is_numeric($k))
			$parent->setAttribute('name', $k);
		if (is_object($v)) {
			$v = get_object_vars($v);
		}
		if (is_array($v)) {
			foreach ($v as $vk => $vv) {
				$value = $doc->createElement($elem_name);
				$parent->appendChild($value);
				$this->xml_encode_worker($vk, $vv, $value, $doc, "value");
			}
		}
		else {
			// regex found at http://stackoverflow.com/a/3466609/49879
			$v = preg_replace('/[^\x{0009}\x{000a}\x{000d}\x{0020}-\x{D7FF}\x{E000}-\x{FFFD}]+/u', '', $v);
			$parent->appendChild($doc->createTextNode($v));
		}
	}

	/**
	 * We need to do the seemingly annoying<key>$key</key><value>$value</value>
	 * encoding, because timeperiod exceptions and variables such as 2dcoords
	 * are not valid as XML keys
	 */
	private function xml_encode($data, $root_name='root') {
		$doc = new DOMDocument();
		$doc->formatOutput = true;
		$root = $doc->createElement($root_name);
		$doc->appendChild($root);
		$this->xml_encode_worker(0, $data, $root, $doc, "param");
		return $doc;
	}

	/**
	 * @param $data mixed
	 * @param $escape boolean = false, not used for the whole body but used when listing objects' properties
	 */
	private function html_encode($data, $escape = false) {
		if(is_bool($data)) {
			return (int) $data;
		}
		if(is_string($data)) {
			if (empty($data))
				return '[empty]';
			if($escape) {
				return htmlspecialchars($data);
			}
			return $data;
		}
		if(is_numeric($data)) {
			return (string) $data;
		}
		$string = null;
		if(is_array($data)) {
			if(empty($data)) {
				return '[empty]';
			}
			if (array_keys($data) !== range(0, count($data) - 1)) {
				// Associative map/hashmap/dictionary (copied from Zend's json encoding)
				$string .= '<dl>';
				foreach($data as $key => $value) {
					if(in_array($key, $this->properties_to_linkup) && is_string($value)) {
						$string .= "<dt>$key</dt><dd><a href='$value'>".htmlspecialchars($value)."</a></dd>";
					} else {
						$string .= "<dt>".htmlspecialchars($key)."</dt><dd>".$this->html_encode($value, true)."</dd>";
					}
				}
				$string .= '</dl>';
			} else {
				// Array/list
				$string .= '<ul>';
				foreach($data as $value) {
					$string .= "<li>".$this->html_encode($value, true)."</li>";
				}
				$string .= '</ul>';
			}
		}
		if($data instanceof stdClass) {
			return $this->html_encode((array) $data);
		}
		return $string;
	}
}
